(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Compare_dec Le Image List Ensembles Mult Program.
Require Export FunctionalExtensionality Classical IndefiniteDescription ChoiceFacts.
Require Import vgtac hpattern.

Set Implicit Arguments.


(** ** Classical logic.

    We assume classical logic as well as functional and propositional
    extensionality.
*)


Lemma prop_ext:
  ClassicalFacts.prop_extensionality.
Proof.
  red; i; exploit (@Extensionality_Ensembles () (fun _ => A) (fun _ => B)).
  - split; red; intros; unfold In in *; intuition.
  - intro EQ; change A with ((fun _ => A) ()).
    by change B with ((fun _ => B) ()); rewrite EQ.
Qed.

Lemma proof_irr:
  ClassicalFacts.proof_irrelevance.
Proof.
  exact (ClassicalFacts.ext_prop_dep_proof_irrel_cic prop_ext).
Qed.

Lemma proj1_sig_inj: forall A (N : A -> Prop) (a a': sig N)
    (EQ: `a = `a'),
  a = a'.
Proof.
  i; destruct a, a'; ss; subst; f_equal; eauto using proof_irr.
Qed.

Lemma sig_eta: forall A (P: A -> Prop) (x : sig P) pf,
  exist P (`x) pf = x.
Proof.
  by i; destruct x; ss; apply proj1_sig_inj.
Qed.

Lemma LEM: forall (P: Prop),
  {P} + {~P}.
Proof.
  i; eapply constructive_definite_descr_excluded_middle; eauto using classic.
  ii; eapply constructive_indefinite_description in H.
  destruct H as (a & H); destruct H; eauto.
Qed.

Lemma or2sum: forall P Q
    (OR: P \/ Q),
  {P} + {Q}.
Proof.
  i; destruct (LEM P); [by left|]; destruct (LEM Q); [by right|].
  exfalso; intuition.
Qed.

Ltac gen_cid :=
  match goal with
    |- context[@constructive_indefinite_description ?A ?B ?C]
    => generalize (@constructive_indefinite_description A B C)
  end.

Ltac gen_eqrefl :=
  match goal with
    |- context[@eq_refl ?A ?B]
    => generalize (@eq_refl A B) 
  end.
