(*********************************************************************)
(**  Pilsner: A Compositionally Verified Compiler for a             **)
(**  Higher-Order Imperative Language                               **)
(**                                                                 **)
(**  For copyright and license, see LICENSE.txt.                    **)
(**  For anything else, see README.txt.                             **)
(*********************************************************************)

Require Import Arith Bool List.
Require Import vgtac basis types lang_common lang_mid.
Require Import model_untyped gw_common gw_untyped.

Set Implicit Arguments.

Global Set Bullet Behavior "Strict Subproofs".


(** IL module transformations. *)

(** The context argument is needed for [opt_contify]. :( *)
Definition modtf :=
  list label -> module -> module.

Definition modtf_wf (f: modtf) := forall FI m FO
    (WF: mid_wfmod FI m FO)
    (UN: uniqvar m),
  <<WF: mid_wfmod FI (f FI m) FO>> /\
  <<UN: uniqvar (f FI m)>>.

Definition modtf_sim (f: modtf) := forall FI m FO
    (WF: mid_wfmod FI m FO)
    (UN: uniqvar m),
  rtc (msim mm_gw FI FO) (f FI m) m.


Definition apply_modtfs (fs: list modtf) : modtf :=
  fun FI => fold_left (fun m f => f FI m) fs.

Lemma apply_modtfs_correct: forall tfs FI FO m
    (WF: lang_mid.mid_wfmod FI m FO)
    (UN: lang_mid.uniqvar m)
    (TF: forall f (IN: In f tfs), <<WF: modtf_wf f>> /\ <<SIM: modtf_sim f>>),
  <<WF: lang_mid.mid_wfmod FI (apply_modtfs tfs FI m) FO>> /\
  <<SIM: rtc (model_untyped.msim mm_gw FI FO) (apply_modtfs tfs FI m) m>>.
Proof.
  induction tfs; [by auto|]; i.
  assert (TF' := TF a (or_introl eq_refl)); des.
  exploit WF0; eauto; []; i; des.
  exploit SIM; eauto; []; i; des.
  hexploit IHtfs; [by eauto|by eauto|by ss; eauto|]; i; des.
  split; [|econs 3]; eauto.
Qed.
